@@ -607,16 +607,20 @@ func (pc *Client) PublishPolicyPack(ctx context.Context, orgName string,
 	}
 
 	//
-	// Step 2: Upload the compressed PolicyPack directory to the presigned S3 URL. The PolicyPack is
-	// now published.
+	// Step 2: Upload the compressed PolicyPack directory to the pre-signed object storage service URL.
+	// The PolicyPack is now published.
 	//
 
-	putS3Req, err := http.NewRequest(http.MethodPut, resp.UploadURI, dirArchive)
+	putReq, err := http.NewRequest(http.MethodPut, resp.UploadURI, dirArchive)
 	if err != nil {
 		return "", errors.Wrapf(err, "Failed to upload compressed PolicyPack")
 	}
 
-	_, err = http.DefaultClient.Do(putS3Req)
+	for k, v := range resp.RequiredHeaders {
+		putReq.Header.Add(k, v)
+	}
+
+	_, err = http.DefaultClient.Do(putReq)
 	if err != nil {
 		return "", errors.Wrapf(err, "Failed to upload compressed PolicyPack")
 	}
