@@ -284,7 +284,7 @@ func TestResourceNestedPropertyPythonCasing(t *testing.T) {
 		}
 
 		t.Run("InputPropertiesAreSnakeCased", func(t *testing.T) {
-			props := mod.getProperties(r.InputProperties, "python", true, true, false, false)
+			props := mod.getProperties(r.InputProperties, "python", true, false, false)
 			for _, p := range props {
 				assert.True(t, strings.Contains(p.Name, "_"), "input property name in python must use snake_case")
 			}
@@ -552,6 +552,15 @@ func TestGeneratePackage(t *testing.T) {
 				"pet.md",
 			},
 		},
+		{
+			"Repro for #6957",
+			"plain-schema-gh6957",
+			[]string{
+				"_index.md",
+				"provider.md",
+				"staticpage.md",
+			},
+		},
 	}
 	testDir := filepath.Join("..", "internal", "test", "testdata")
 	for _, tt := range tests {
