@@ -25,6 +25,7 @@ type generator struct {
 	packages            map[string]*schema.Package
 	contexts            map[string]map[string]*pkgContext
 	diagnostics         hcl.Diagnostics
+	spills              *spills
 	jsonTempSpiller     *jsonSpiller
 	ternaryTempSpiller  *tempSpiller
 	readDirTempSpiller  *readDirSpiller
@@ -49,6 +50,7 @@ func GenerateProgram(program *hcl2.Program) (map[string][]byte, hcl.Diagnostics,
 		program:             program,
 		packages:            packages,
 		contexts:            contexts,
+		spills:              &spills{counts: map[string]int{}},
 		jsonTempSpiller:     &jsonSpiller{},
 		ternaryTempSpiller:  &tempSpiller{},
 		readDirTempSpiller:  &readDirSpiller{},
@@ -142,6 +144,48 @@ func (g *generator) genPreamble(w io.Writer, program *hcl2.Program, stdImports,
 	g.Fprintf(w, "pulumi.Run(func(ctx *pulumi.Context) error {\n")
 }
 
+func (g *generator) collectTypeImports(program *hcl2.Program, t schema.Type, imports codegen.StringSet) {
+	var token string
+	switch t := t.(type) {
+	case *schema.InputType:
+		g.collectTypeImports(program, t.ElementType, imports)
+		return
+	case *schema.OptionalType:
+		g.collectTypeImports(program, t.ElementType, imports)
+		return
+	case *schema.ArrayType:
+		g.collectTypeImports(program, t.ElementType, imports)
+		return
+	case *schema.MapType:
+		g.collectTypeImports(program, t.ElementType, imports)
+		return
+	case *schema.UnionType:
+		for _, t := range t.ElementTypes {
+			g.collectTypeImports(program, t, imports)
+		}
+		return
+	case *schema.ObjectType:
+		token = t.Token
+	case *schema.EnumType:
+		token = t.Token
+	case *schema.TokenType:
+		token = t.Token
+	case *schema.ResourceType:
+		token = t.Token
+	}
+	if token == "" {
+		return
+	}
+
+	var tokenRange hcl.Range
+	pkg, mod, _, _ := hcl2.DecomposeToken(token, tokenRange)
+	vPath, err := g.getVersionPath(program, pkg)
+	if err != nil {
+		panic(err)
+	}
+	imports.Add(g.getPulumiImport(pkg, vPath, mod))
+}
+
 // collect Imports returns two sets of packages imported by the program, std lib packages and pulumi packages
 func (g *generator) collectImports(
 	program *hcl2.Program,
@@ -183,27 +227,7 @@ func (g *generator) collectImports(
 					pulumiImports.Add(g.getPulumiImport(pkg, vPath, mod))
 				} else if call.Name == hcl2.IntrinsicConvert {
 					if schemaType, ok := hcl2.GetSchemaForType(call.Type()); ok {
-						switch schemaType := schemaType.(type) {
-						case *schema.ObjectType:
-							token := schemaType.Token
-							var tokenRange hcl.Range
-							pkg, mod, _, _ := hcl2.DecomposeToken(token, tokenRange)
-							vPath, err := g.getVersionPath(program, pkg)
-							if err != nil {
-								panic(err)
-							}
-							pulumiImports.Add(g.getPulumiImport(pkg, vPath, mod))
-						case *schema.ArrayType:
-							token := schemaType.ElementType.(*schema.ObjectType).Token
-							var tokenRange hcl.Range
-							pkg, mod, _, _ := hcl2.DecomposeToken(token, tokenRange)
-							vPath, err := g.getVersionPath(program, pkg)
-							if err != nil {
-								panic(err)
-							}
-							pulumiImports.Add(g.getPulumiImport(pkg, vPath, mod))
-						}
-
+						g.collectTypeImports(program, schemaType, pulumiImports)
 					}
 				}
 			}
@@ -342,7 +366,7 @@ func (g *generator) lowerResourceOptions(opts *hcl2.ResourceOptions) (*model.Blo
 			}
 		}
 
-		value, valueTemps := g.lowerExpression(value, destType, false)
+		value, valueTemps := g.lowerExpression(value, destType)
 		temps = append(temps, valueTemps...)
 
 		block.Body.Items = append(block.Body.Items, &model.Attribute{
@@ -398,8 +422,7 @@ func (g *generator) genResource(w io.Writer, r *hcl2.Resource) {
 	for _, input := range r.Inputs {
 		destType, diagnostics := r.InputType.Traverse(hcl.TraverseAttr{Name: input.Name})
 		g.diagnostics = append(g.diagnostics, diagnostics...)
-		isInput := true
-		expr, temps := g.lowerExpression(input.Value, destType.(model.Type), isInput)
+		expr, temps := g.lowerExpression(input.Value, destType.(model.Type))
 		input.Value = expr
 		g.genTemps(w, temps)
 	}
@@ -438,7 +461,7 @@ func (g *generator) genResource(w io.Writer, r *hcl2.Resource) {
 
 	if r.Options != nil && r.Options.Range != nil {
 		rangeType := model.ResolveOutputs(r.Options.Range.Type())
-		rangeExpr, temps := g.lowerExpression(r.Options.Range, rangeType, false)
+		rangeExpr, temps := g.lowerExpression(r.Options.Range, rangeType)
 		g.genTemps(w, temps)
 
 		g.Fgenf(w, "var %s []*%s.%s\n", resName, modOrAlias, typ)
@@ -466,8 +489,7 @@ func (g *generator) genResource(w io.Writer, r *hcl2.Resource) {
 }
 
 func (g *generator) genOutputAssignment(w io.Writer, v *hcl2.OutputVariable) {
-	isInput := false
-	expr, temps := g.lowerExpression(v.Value, v.Type(), isInput)
+	expr, temps := g.lowerExpression(v.Value, v.Type())
 	g.genTemps(w, temps)
 	g.Fgenf(w, "ctx.Export(\"%s\", %.3v)\n", v.Name(), expr)
 }
@@ -482,7 +504,7 @@ func (g *generator) genTempsMultiReturn(w io.Writer, temps []interface{}, zeroVa
 	if zeroValueType != "" {
 		for _, t := range temps {
 			switch t.(type) {
-			case *jsonTemp, *readDirTemp:
+			case *spillTemp, *jsonTemp, *readDirTemp:
 				genZeroValueDecl = true
 			default:
 			}
@@ -506,10 +528,10 @@ func (g *generator) genTempsMultiReturn(w io.Writer, temps []interface{}, zeroVa
 			g.Fgenf(w, "} else {\n")
 			g.Fgenf(w, "%s = %.v\n", t.Name, t.Value.FalseResult)
 			g.Fgenf(w, "}\n")
-		case *jsonTemp:
-			bytesVar := fmt.Sprintf("tmp%s", strings.ToUpper(t.Name))
+		case *spillTemp:
+			bytesVar := fmt.Sprintf("tmp%s", strings.ToUpper(t.Variable.Name))
 			g.Fgenf(w, "%s, err := json.Marshal(", bytesVar)
-			args := stripInputs(t.Value.Args[0])
+			args := t.Value.(*model.FunctionCallExpression).Args[0]
 			g.Fgenf(w, "%.v)\n", args)
 			g.Fgenf(w, "if err != nil {\n")
 			if genZeroValueDecl {
@@ -518,7 +540,7 @@ func (g *generator) genTempsMultiReturn(w io.Writer, temps []interface{}, zeroVa
 				g.Fgenf(w, "return err\n")
 			}
 			g.Fgenf(w, "}\n")
-			g.Fgenf(w, "%s := string(%s)\n", t.Name, bytesVar)
+			g.Fgenf(w, "%s := string(%s)\n", t.Variable.Name, bytesVar)
 		case *readDirTemp:
 			tmpSuffix := strings.Split(t.Name, "files")[1]
 			g.Fgenf(w, "%s, err := ioutil.ReadDir(%.v)\n", t.Name, t.Value.Args[0])
@@ -555,8 +577,7 @@ func (g *generator) genTempsMultiReturn(w io.Writer, temps []interface{}, zeroVa
 }
 
 func (g *generator) genLocalVariable(w io.Writer, v *hcl2.LocalVariable) {
-	isInput := false
-	expr, temps := g.lowerExpression(v.Definition.Value, v.Type(), isInput)
+	expr, temps := g.lowerExpression(v.Definition.Value, v.Type())
 	g.genTemps(w, temps)
 	name := makeValidIdentifier(v.Name())
 	assignment := ":="
@@ -609,7 +630,7 @@ func (g *generator) genConfigVariable(w io.Writer, v *hcl2.ConfigVariable) {
 	if v.DefaultValue == nil {
 		g.Fgenf(w, "%[1]s := cfg.%[2]s%[3]s(\"%[1]s\")\n", v.Name(), getOrRequire, getType)
 	} else {
-		expr, temps := g.lowerExpression(v.DefaultValue, v.DefaultValue.Type(), false)
+		expr, temps := g.lowerExpression(v.DefaultValue, v.DefaultValue.Type())
 		g.genTemps(w, temps)
 		switch expr := expr.(type) {
 		case *model.FunctionCallExpression:
