@@ -116,6 +116,7 @@ func TestGeneratePackage(t *testing.T) {
 				filepath.Join("example", "doc.go"),
 				filepath.Join("example", "init.go"),
 				filepath.Join("example", "component.go"),
+				filepath.Join("example", "doFoo.go"),
 				filepath.Join("example", "provider.go"),
 				filepath.Join("example", "pulumiTypes.go"),
 				filepath.Join("example", "pulumiUtilities.go"),
@@ -135,6 +136,19 @@ func TestGeneratePackage(t *testing.T) {
 			},
 			false,
 		},
+		{
+			"Repro for #6957",
+			"plain-schema-gh6957",
+			[]string{
+				filepath.Join("xyz", "doc.go"),
+				filepath.Join("xyz", "init.go"),
+				filepath.Join("xyz", "staticPage.go"),
+				filepath.Join("xyz", "provider.go"),
+				filepath.Join("xyz", "pulumiTypes.go"),
+				filepath.Join("xyz", "pulumiUtilities.go"),
+			},
+			false,
+		},
 	}
 	testDir := filepath.Join("..", "internal", "test", "testdata")
 	for _, tt := range tests {
