@@ -84,7 +84,7 @@ func (primitiveType) isType() {}
 // IsPrimitiveType returns true if the given Type is a primitive type. The primitive types are bool, int, number,
 // string, archive, asset, and any.
 func IsPrimitiveType(t Type) bool {
-	_, ok := t.(primitiveType)
+	_, ok := plainType(t).(primitiveType)
 	return ok
 }
 
@@ -133,6 +133,8 @@ func (*ArrayType) isType() {}
 
 // EnumType represents an enum.
 type EnumType struct {
+	// Package is the type's package.
+	Package *Package
 	// Token is the type's Pulumi type token.
 	Token string
 	// Comment is the description of the type, if any.
@@ -204,9 +206,28 @@ type ObjectType struct {
 	// Language specifies additional language-specific data about the object type.
 	Language map[string]interface{}
 
+	// InputShape is the input shape for this object. Only valid if IsPlainShape returns true.
+	InputShape *ObjectType
+	// PlainShape is the plain shape for this object. Only valid if IsInputShape returns true.
+	PlainShape *ObjectType
+
 	properties map[string]*Property
 }
 
+// IsPlainShape returns true if this object type is the plain shape of a (plain, input)
+// pair. The plain shape of an object does not contain *InputType values and only
+// references other plain shapes.
+func (t *ObjectType) IsPlainShape() bool {
+	return t.PlainShape == nil
+}
+
+// IsInputShape returns true if this object type is the plain shape of a (plain, input)
+// pair. The input shape of an object may contain *InputType values and may
+// reference other input shapes.
+func (t *ObjectType) IsInputShape() bool {
+	return t.PlainShape != nil
+}
+
 func (t *ObjectType) Property(name string) (*Property, bool) {
 	if t.properties == nil && len(t.Properties) > 0 {
 		t.properties = make(map[string]*Property)
@@ -219,6 +240,9 @@ func (t *ObjectType) Property(name string) (*Property, bool) {
 }
 
 func (t *ObjectType) String() string {
+	if t.PlainShape != nil {
+		return t.Token + "•Input"
+	}
 	return t.Token
 }
 
@@ -252,6 +276,30 @@ func (t *TokenType) String() string {
 
 func (*TokenType) isType() {}
 
+// InputType represents a type that accepts either a prompt value or an output value.
+type InputType struct {
+	// ElementType is the element type of the input.
+	ElementType Type
+}
+
+func (t *InputType) String() string {
+	return fmt.Sprintf("Input<%v>", t.ElementType)
+}
+
+func (*InputType) isType() {}
+
+// OptionalType represents a type that accepts an optional value.
+type OptionalType struct {
+	// ElementType is the element type of the input.
+	ElementType Type
+}
+
+func (t *OptionalType) String() string {
+	return fmt.Sprintf("Optional<%v>", t.ElementType)
+}
+
+func (*OptionalType) isType() {}
+
 // DefaultValue describes a default value for a property.
 type DefaultValue struct {
 	// Value specifies a static default value, if any. This value must be representable in the Pulumi schema type
@@ -275,10 +323,6 @@ type Property struct {
 	ConstValue interface{}
 	// DefaultValue is the default value for the property, if any.
 	DefaultValue *DefaultValue
-	// IsRequired is true if the property must always be populated.
-	IsRequired bool
-	// IsPlain is true if the property only accepts prompt values.
-	IsPlain bool
 	// DeprecationMessage indicates whether or not the property is deprecated.
 	DeprecationMessage string
 	// Language specifies additional language-specific data about the property.
@@ -287,6 +331,12 @@ type Property struct {
 	Secret bool
 }
 
+// IsRequired returns true if this property is required (i.e. its type is not Optional).
+func (p *Property) IsRequired() bool {
+	_, optional := p.Type.(*OptionalType)
+	return !optional
+}
+
 // Alias describes an alias for a Pulumi resource.
 type Alias struct {
 	// Name is the "name" portion of the alias, if any.
@@ -669,6 +719,8 @@ type TypeSpec struct {
 	OneOf []TypeSpec `json:"oneOf,omitempty"`
 	// Discriminator informs the consumer of an alternative schema based on the value associated with it.
 	Discriminator *DiscriminatorSpec `json:"discriminator,omitempty"`
+	// Plain indicates that when used as an input, this type does not accept eventual values.
+	Plain bool `json:"plain,omitempty"`
 }
 
 // DiscriminatorSpec informs the consumer of an alternative schema based on the value associated with it.
@@ -720,8 +772,8 @@ type ObjectTypeSpec struct {
 	// Required, if present, is a list of the names of an object type's required properties. These properties must be set
 	// for inputs and will always be set for outputs.
 	Required []string `json:"required,omitempty"`
-	// Plain, if present, is a list of the names of an object type's plain properties. These properties only accept
-	// prompt values.
+	// Plain, was a list of the names of an object type's plain properties. This property is ignored: instead, property
+	// types should be marked as plain where necessary.
 	Plain []string `json:"plain,omitempty"`
 	// Language specifies additional language-specific data about the type.
 	Language map[string]json.RawMessage `json:"language,omitempty"`
@@ -765,7 +817,8 @@ type ResourceSpec struct {
 	InputProperties map[string]PropertySpec `json:"inputProperties,omitempty"`
 	// RequiredInputs is a list of the names of the resource's required input properties.
 	RequiredInputs []string `json:"requiredInputs,omitempty"`
-	// PlainInputs is a list of the names of the resource's plain input properties that only accept prompt values.
+	// PlainInputs was a list of the names of the resource's plain input properties. This property is ignored:
+	// instead, property types should be marked as plain where necessary.
 	PlainInputs []string `json:"plainInputs,omitempty"`
 	// StateInputs is an optional ObjectTypeSpec that describes additional inputs that mau be necessary to get an
 	// existing resource. If this is unset, only an ID is necessary.
@@ -927,7 +980,9 @@ func importSpec(spec PackageSpec, languages map[string]Language, loader Loader)
 		typeList = append(typeList, t)
 	}
 	for _, t := range types.objects {
+		// t is a plain shape: add it and its coresponding input shape to the type list.
 		typeList = append(typeList, t)
+		typeList = append(typeList, t.InputShape)
 	}
 	for _, t := range types.arrays {
 		typeList = append(typeList, t)
@@ -1003,6 +1058,8 @@ type types struct {
 	tokens    map[string]*TokenType
 	enums     map[string]*EnumType
 	named     map[string]Type // objects and enums
+	inputs    map[Type]*InputType
+	optionals map[Type]*OptionalType
 }
 
 func (t *types) bindPrimitiveType(name string) (Type, error) {
@@ -1120,7 +1177,65 @@ func versionEquals(a, b *semver.Version) bool {
 	return a.Equals(*b)
 }
 
-func (t *types) bindTypeSpecRef(spec TypeSpec) (Type, error) {
+func (t *types) newInputType(elementType Type) Type {
+	if _, ok := elementType.(*InputType); ok {
+		return elementType
+	}
+
+	typ, ok := t.inputs[elementType]
+	if !ok {
+		typ = &InputType{ElementType: elementType}
+		t.inputs[elementType] = typ
+	}
+	return typ
+}
+
+func (t *types) newOptionalType(elementType Type) Type {
+	if _, ok := elementType.(*OptionalType); ok {
+		return elementType
+	}
+	typ, ok := t.optionals[elementType]
+	if !ok {
+		typ = &OptionalType{ElementType: elementType}
+		t.optionals[elementType] = typ
+	}
+	return typ
+}
+
+func (t *types) newMapType(elementType Type) Type {
+	typ, ok := t.maps[elementType]
+	if !ok {
+		typ = &MapType{ElementType: elementType}
+		t.maps[elementType] = typ
+	}
+	return typ
+}
+
+func (t *types) newArrayType(elementType Type) Type {
+	typ, ok := t.arrays[elementType]
+	if !ok {
+		typ = &ArrayType{ElementType: elementType}
+		t.arrays[elementType] = typ
+	}
+	return typ
+}
+
+func (t *types) newUnionType(
+	elements []Type, defaultType Type, discriminator string, mapping map[string]string) *UnionType {
+	union := &UnionType{
+		ElementTypes:  elements,
+		DefaultType:   defaultType,
+		Discriminator: discriminator,
+		Mapping:       mapping,
+	}
+	if typ, ok := t.unions[union.String()]; ok {
+		return typ
+	}
+	t.unions[union.String()] = union
+	return union
+}
+
+func (t *types) bindTypeSpecRef(spec TypeSpec, inputShape bool) (Type, error) {
 	// Explicitly handle built-in types so that we don't have to handle this type of path during ref parsing.
 	switch spec.Ref {
 	case "pulumi.json#/Archive":
@@ -1152,6 +1267,9 @@ func (t *types) bindTypeSpecRef(spec TypeSpec) (Type, error) {
 			if !ok {
 				return nil, fmt.Errorf("type %v not found in package %v", ref.Token, ref.Package)
 			}
+			if obj, ok := typ.(*ObjectType); ok && inputShape {
+				typ = obj.InputShape
+			}
 			return typ, nil
 		case resourcesRef, providerRef:
 			typ, ok := pkg.GetResourceType(ref.Token)
@@ -1165,6 +1283,9 @@ func (t *types) bindTypeSpecRef(spec TypeSpec) (Type, error) {
 	switch ref.Kind {
 	case typesRef:
 		if typ, ok := t.objects[ref.Token]; ok {
+			if inputShape {
+				return typ.InputShape, nil
+			}
 			return typ, nil
 		}
 		if typ, ok := t.enums[ref.Token]; ok {
@@ -1174,7 +1295,7 @@ func (t *types) bindTypeSpecRef(spec TypeSpec) (Type, error) {
 		if !ok {
 			typ = &TokenType{Token: ref.Token}
 			if spec.Type != "" {
-				ut, err := t.bindType(TypeSpec{Type: spec.Type})
+				ut, err := t.bindPrimitiveType(spec.Type)
 				if err != nil {
 					return nil, err
 				}
@@ -1195,9 +1316,15 @@ func (t *types) bindTypeSpecRef(spec TypeSpec) (Type, error) {
 	}
 }
 
-func (t *types) bindType(spec TypeSpec) (Type, error) {
+func (t *types) bindType(spec TypeSpec, inputShape bool) (result Type, err error) {
+	if inputShape && !spec.Plain {
+		defer func() {
+			result = t.newInputType(result)
+		}()
+	}
+
 	if spec.Ref != "" {
-		return t.bindTypeSpecRef(spec)
+		return t.bindTypeSpecRef(spec, inputShape)
 	}
 
 	if spec.OneOf != nil {
@@ -1216,7 +1343,7 @@ func (t *types) bindType(spec TypeSpec) (Type, error) {
 
 		elements := make([]Type, len(spec.OneOf))
 		for i, spec := range spec.OneOf {
-			e, err := t.bindType(spec)
+			e, err := t.bindType(spec, inputShape)
 			if err != nil {
 				return nil, err
 			}
@@ -1230,17 +1357,7 @@ func (t *types) bindType(spec TypeSpec) (Type, error) {
 			mapping = spec.Discriminator.Mapping
 		}
 
-		union := &UnionType{
-			ElementTypes:  elements,
-			DefaultType:   defaultType,
-			Discriminator: discriminator,
-			Mapping:       mapping,
-		}
-		if typ, ok := t.unions[union.String()]; ok {
-			return typ, nil
-		}
-		t.unions[union.String()] = union
-		return union, nil
+		return t.newUnionType(elements, defaultType, discriminator, mapping), nil
 	}
 
 	// nolint: goconst
@@ -1252,44 +1369,52 @@ func (t *types) bindType(spec TypeSpec) (Type, error) {
 			return nil, errors.Errorf("missing \"items\" property in type spec")
 		}
 
-		elementType, err := t.bindType(*spec.Items)
+		elementType, err := t.bindType(*spec.Items, inputShape)
 		if err != nil {
 			return nil, err
 		}
 
-		typ, ok := t.arrays[elementType]
-		if !ok {
-			typ = &ArrayType{ElementType: elementType}
-			t.arrays[elementType] = typ
-		}
-		return typ, nil
+		return t.newArrayType(elementType), nil
 	case "object":
 		elementType := StringType
 		if spec.AdditionalProperties != nil {
-			et, err := t.bindType(*spec.AdditionalProperties)
+			et, err := t.bindType(*spec.AdditionalProperties, inputShape)
 			if err != nil {
 				return nil, err
 			}
 			elementType = et
 		}
 
-		typ, ok := t.maps[elementType]
-		if !ok {
-			typ = &MapType{ElementType: elementType}
-			t.maps[elementType] = typ
-		}
-		return typ, nil
+		return t.newMapType(elementType), nil
 	default:
 		return nil, errors.Errorf("unknown type kind %v", spec.Type)
 	}
 }
 
+func plainType(typ Type) Type {
+	for {
+		switch t := typ.(type) {
+		case *InputType:
+			typ = t.ElementType
+		case *OptionalType:
+			typ = t.ElementType
+		case *ObjectType:
+			if t.PlainShape == nil {
+				return t
+			}
+			typ = t.PlainShape
+		default:
+			return t
+		}
+	}
+}
+
 func bindConstValue(value interface{}, typ Type) (interface{}, error) {
 	if value == nil {
 		return nil, nil
 	}
 
-	switch typ {
+	switch plainType(typ) {
 	case BoolType:
 		if _, ok := value.(bool); !ok {
 			return nil, errors.Errorf("invalid constant of type %T for boolean property", value)
@@ -1324,6 +1449,7 @@ func bindDefaultValue(value interface{}, spec *DefaultSpec, typ Type) (*DefaultV
 	}
 
 	if value != nil {
+		typ = plainType(typ)
 		switch typ := typ.(type) {
 		case *UnionType:
 			if typ.DefaultType != nil {
@@ -1380,14 +1506,14 @@ func bindDefaultValue(value interface{}, spec *DefaultSpec, typ Type) (*DefaultV
 
 // bindProperties binds the map of property specs and list of required properties into a sorted list of properties and
 // a lookup table.
-func (t *types) bindProperties(properties map[string]PropertySpec, required []string,
-	plain []string) ([]*Property, map[string]*Property, error) {
+func (t *types) bindProperties(properties map[string]PropertySpec,
+	required []string, inputShape bool) ([]*Property, map[string]*Property, error) {
 
 	// Bind property types and constant or default values.
 	propertyMap := map[string]*Property{}
 	var result []*Property
 	for name, spec := range properties {
-		typ, err := t.bindType(spec.TypeSpec)
+		typ, err := t.bindType(spec.TypeSpec, inputShape)
 		if err != nil {
 			return nil, nil, errors.Wrapf(err, "error binding type for property %q", name)
 		}
@@ -1410,7 +1536,7 @@ func (t *types) bindProperties(properties map[string]PropertySpec, required []st
 		p := &Property{
 			Name:               name,
 			Comment:            spec.Description,
-			Type:               typ,
+			Type:               t.newOptionalType(typ),
 			ConstValue:         cv,
 			DefaultValue:       dv,
 			DeprecationMessage: spec.DeprecationMessage,
@@ -1427,16 +1553,7 @@ func (t *types) bindProperties(properties map[string]PropertySpec, required []st
 		if !ok {
 			return nil, nil, errors.Errorf("unknown required property %q", name)
 		}
-		p.IsRequired = true
-	}
-
-	// Compute plain properties.
-	for _, name := range plain {
-		p, ok := propertyMap[name]
-		if !ok {
-			return nil, nil, errors.Errorf("unknown plain property %q", name)
-		}
-		p.IsPlain = true
+		p.Type = p.Type.(*OptionalType).ElementType
 	}
 
 	sort.Slice(result, func(i, j int) bool {
@@ -1446,14 +1563,17 @@ func (t *types) bindProperties(properties map[string]PropertySpec, required []st
 	return result, propertyMap, nil
 }
 
-func (t *types) bindObjectTypeDetails(obj *ObjectType, token string, spec ObjectTypeSpec,
-	supportsPlainProperties bool) error {
-
-	if !supportsPlainProperties && len(spec.Plain) > 0 {
-		return errors.New("plain cannot be specified")
+func (t *types) bindObjectTypeDetails(obj *ObjectType, token string, spec ObjectTypeSpec) error {
+	if len(spec.Plain) > 0 {
+		return errors.New("plain has been removed; the property type must be marked as plain instead")
 	}
 
-	properties, propertyMap, err := t.bindProperties(spec.Properties, spec.Required, spec.Plain)
+	properties, propertyMap, err := t.bindProperties(spec.Properties, spec.Required, false)
+	if err != nil {
+		return err
+	}
+
+	inputProperties, inputPropertyMap, err := t.bindProperties(spec.Properties, spec.Required, true)
 	if err != nil {
 		return err
 	}
@@ -1469,12 +1589,21 @@ func (t *types) bindObjectTypeDetails(obj *ObjectType, token string, spec Object
 	obj.Language = language
 	obj.Properties = properties
 	obj.properties = propertyMap
+
+	obj.InputShape.Package = t.pkg
+	obj.InputShape.Token = token
+	obj.InputShape.Comment = spec.Description
+	obj.InputShape.Language = language
+	obj.InputShape.Properties = inputProperties
+	obj.InputShape.properties = inputPropertyMap
+
 	return nil
 }
 
-func (t *types) bindObjectType(token string, spec ObjectTypeSpec, supportsPlainProperties bool) (*ObjectType, error) {
+func (t *types) bindObjectType(token string, spec ObjectTypeSpec) (*ObjectType, error) {
 	obj := &ObjectType{}
-	if err := t.bindObjectTypeDetails(obj, token, spec, supportsPlainProperties); err != nil {
+	obj.InputShape = &ObjectType{PlainShape: obj}
+	if err := t.bindObjectTypeDetails(obj, token, spec); err != nil {
 		return nil, err
 	}
 	return obj, nil
@@ -1494,7 +1623,7 @@ func (t *types) bindResourceType(token string) (*ResourceType, error) {
 }
 
 func (t *types) bindEnumTypeDetails(enum *EnumType, token string, spec ComplexTypeSpec) error {
-	typ, err := t.bindType(TypeSpec{Type: spec.Type})
+	typ, err := t.bindPrimitiveType(spec.Type)
 	if err != nil {
 		return err
 	}
@@ -1504,6 +1633,7 @@ func (t *types) bindEnumTypeDetails(enum *EnumType, token string, spec ComplexTy
 		return err
 	}
 
+	enum.Package = t.pkg
 	enum.Token = token
 	enum.Elements = values
 	enum.ElementType = typ
@@ -1564,7 +1694,6 @@ func (t *types) bindEnumType(token string, spec ComplexTypeSpec) (*EnumType, err
 }
 
 func bindTypes(pkg *Package, complexTypes map[string]ComplexTypeSpec, loader Loader) (*types, error) {
-
 	typs := &types{
 		pkg:       pkg,
 		loader:    loader,
@@ -1576,6 +1705,8 @@ func bindTypes(pkg *Package, complexTypes map[string]ComplexTypeSpec, loader Loa
 		tokens:    map[string]*TokenType{},
 		enums:     map[string]*EnumType{},
 		named:     map[string]Type{},
+		inputs:    map[Type]*InputType{},
+		optionals: map[Type]*OptionalType{},
 	}
 
 	// Declare object and enum types before processing properties.
@@ -1586,6 +1717,7 @@ func bindTypes(pkg *Package, complexTypes map[string]ComplexTypeSpec, loader Loa
 			// object type is its token. While this doesn't affect object types directly, it breaks the interning of types
 			// that reference object types (e.g. arrays, maps, unions)
 			typ := &ObjectType{Token: token}
+			typ.InputShape = &ObjectType{Token: token, PlainShape: typ}
 			typs.objects[token] = typ
 			typs.named[token] = typ
 		} else if len(spec.Enum) > 0 {
@@ -1609,7 +1741,7 @@ func bindTypes(pkg *Package, complexTypes map[string]ComplexTypeSpec, loader Loa
 	for token, spec := range complexTypes {
 		if spec.Type == "object" {
 			if err := typs.bindObjectTypeDetails(
-				typs.objects[token], token, spec.ObjectTypeSpec, true /*supportsPlainProperties*/); err != nil {
+				typs.objects[token], token, spec.ObjectTypeSpec); err != nil {
 				return nil, errors.Wrapf(err, "failed to bind type %s", token)
 			}
 		}
@@ -1655,25 +1787,25 @@ func bindMethods(resourceToken string, methods map[string]string,
 }
 
 func bindConfig(spec ConfigSpec, types *types) ([]*Property, error) {
-	properties, _, err := types.bindProperties(spec.Variables, spec.Required, nil)
+	properties, _, err := types.bindProperties(spec.Variables, spec.Required, false)
 	return properties, err
 }
 
 func bindResource(token string, spec ResourceSpec, types *types,
 	functionTable map[string]*Function) (*Resource, error) {
 	if len(spec.Plain) > 0 {
-		return nil, errors.New("plain cannot be specified on resources")
+		return nil, errors.New("plain has been removed; property types must be marked as plain instead")
 	}
-	if len(spec.PlainInputs) > 0 && !spec.IsComponent {
-		return nil, errors.New("plainInputs can only be specified on component resources")
+	if len(spec.PlainInputs) > 0 {
+		return nil, errors.New("plainInputs has been removed; individual property types must be marked as plain instead")
 	}
 
-	properties, _, err := types.bindProperties(spec.Properties, spec.Required, nil)
+	properties, _, err := types.bindProperties(spec.Properties, spec.Required, false)
 	if err != nil {
 		return nil, errors.Wrap(err, "failed to bind properties")
 	}
 
-	inputProperties, _, err := types.bindProperties(spec.InputProperties, spec.RequiredInputs, spec.PlainInputs)
+	inputProperties, _, err := types.bindProperties(spec.InputProperties, spec.RequiredInputs, true)
 	if err != nil {
 		return nil, errors.Wrap(err, "failed to bind properties")
 	}
@@ -1691,11 +1823,11 @@ func bindResource(token string, spec ResourceSpec, types *types,
 
 	var stateInputs *ObjectType
 	if spec.StateInputs != nil {
-		si, err := types.bindObjectType(token+"Args", *spec.StateInputs, false /*supportsPlainProperties*/)
+		si, err := types.bindObjectType(token+"Args", *spec.StateInputs)
 		if err != nil {
 			return nil, errors.Wrap(err, "error binding inputs")
 		}
-		stateInputs = si
+		stateInputs = si.InputShape
 	}
 
 	var aliases []*Alias
@@ -1737,12 +1869,13 @@ func bindProvider(pkgName string, spec ResourceSpec, types *types,
 	// strings, or types with an underlying type of string, before we generate the provider code.
 	var stringProperties []*Property
 	for _, prop := range res.Properties {
-		if tokenType, isTokenType := prop.Type.(*TokenType); isTokenType {
+		typ := plainType(prop.Type)
+		if tokenType, isTokenType := typ.(*TokenType); isTokenType {
 			if tokenType.UnderlyingType != stringType {
 				continue
 			}
 		} else {
-			if prop.Type != stringType {
+			if typ != stringType {
 				continue
 			}
 		}
@@ -1794,7 +1927,7 @@ func bindResources(specs map[string]ResourceSpec, types *types,
 func bindFunction(token string, spec FunctionSpec, types *types) (*Function, error) {
 	var inputs *ObjectType
 	if spec.Inputs != nil {
-		ins, err := types.bindObjectType(token+"Args", *spec.Inputs, false /*supportsPlainProperties*/)
+		ins, err := types.bindObjectType(token+"Args", *spec.Inputs)
 		if err != nil {
 			return nil, errors.Wrap(err, "error binding inputs")
 		}
@@ -1803,7 +1936,7 @@ func bindFunction(token string, spec FunctionSpec, types *types) (*Function, err
 
 	var outputs *ObjectType
 	if spec.Outputs != nil {
-		outs, err := types.bindObjectType(token+"Result", *spec.Outputs, false /*supportsPlainProperties*/)
+		outs, err := types.bindObjectType(token+"Result", *spec.Outputs)
 		if err != nil {
 			return nil, errors.Wrap(err, "error binding inputs")
 		}
