@@ -5,10 +5,13 @@ using System.Collections.Concurrent;
 using System.Diagnostics.CodeAnalysis;
 using System.Threading;
 using System.Threading.Tasks;
-using Microsoft.Extensions.Configuration;
+using Microsoft.Extensions.Logging;
 using Pulumi.Testing;
+using Pulumirpc;
 using Serilog;
 using Serilog.Events;
+using Serilog.Extensions.Logging;
+using ILogger = Microsoft.Extensions.Logging.ILogger;
 
 namespace Pulumi
 {
@@ -30,7 +33,7 @@ namespace Pulumi
     /// </para>
     /// Importantly: Cloud resources cannot be created outside of the lambda passed to any of the
     /// <see cref="Deployment.RunAsync(Action)"/> overloads.  Because cloud Resource construction is
-    /// inherently asynchronous, the result of this function is a <see cref="Task{T}"/> which should
+    /// inherently asynchronous, the result of this function is a <see cref="Task{TResult}"/> which should
     /// then be returned or awaited.  This will ensure that any problems that are encountered during
     /// the running of the program are properly reported.  Failure to do this may lead to the
     /// program ending early before all resources are properly registered.
@@ -38,7 +41,7 @@ namespace Pulumi
     public sealed partial class Deployment : IDeploymentInternal
     {
         private static readonly object _instanceLock = new object();
-        private static AsyncLocal<DeploymentInstance?> _instance = new AsyncLocal<DeploymentInstance?>();
+        private static readonly AsyncLocal<DeploymentInstance?> _instance = new AsyncLocal<DeploymentInstance?>();
 
         /// <summary>
         /// The current running deployment instance. This is only available from inside the function
@@ -85,8 +88,7 @@ namespace Pulumi
         private readonly bool _isDryRun;
         private readonly ConcurrentDictionary<string, bool> _featureSupport = new ConcurrentDictionary<string, bool>();
 
-        private Serilog.ILogger _serilogger = null!;
-        private readonly ILogger _logger;
+        private readonly IEngineLogger _logger;
         private readonly IRunner _runner;
 
         internal IEngine Engine { get; }
@@ -101,6 +103,7 @@ namespace Pulumi
 
         private Deployment()
         {
+            // ReSharper disable UnusedVariable
             var monitor = Environment.GetEnvironmentVariable("PULUMI_MONITOR");
             var engine = Environment.GetEnvironmentVariable("PULUMI_ENGINE");
             var project = Environment.GetEnvironmentVariable("PULUMI_PROJECT");
@@ -121,23 +124,24 @@ namespace Pulumi
             {
                 throw new InvalidOperationException("Program run without the Pulumi engine available; re-run using the `pulumi` CLI");
             }
+            // ReSharper restore UnusedVariable
 
             _isDryRun = dryRunValue;
             _stackName = stack;
             _projectName = project;
 
-            InitSerilogger();
+            var deploymentLogger = CreateDefaultLogger();
 
-            _serilogger.Debug("Creating Deployment Engine.");
+            deploymentLogger.LogDebug("Creating deployment engine");
             this.Engine = new GrpcEngine(engine);
-            _serilogger.Debug("Created Deployment Engine.");
+            deploymentLogger.LogDebug("Created deployment engine");
 
-            _serilogger.Debug("Creating Deployment Monitor.");
+            deploymentLogger.LogDebug("Creating deployment monitor");
             this.Monitor = new GrpcMonitor(monitor);
-            _serilogger.Debug("Created Deployment Monitor.");
+            deploymentLogger.LogDebug("Created deployment monitor");
 
-            _runner = new Runner(this);
-            _logger = new Logger(this, this.Engine);
+            _runner = new Runner(this, deploymentLogger);
+            _logger = new EngineLogger(this, deploymentLogger, this.Engine);
         }
 
         /// <summary>
@@ -149,22 +153,21 @@ namespace Pulumi
         /// </summary>
         internal Deployment(IEngine engine, IMonitor monitor, TestOptions? options)
         {
-            InitSerilogger();
+            var deploymentLogger = CreateDefaultLogger();
             _isDryRun = options?.IsPreview ?? true;
             _stackName = options?.StackName ?? "stack";
             _projectName = options?.ProjectName ?? "project";
             this.Engine = engine;
             this.Monitor = monitor;
-            _runner = new Runner(this);
-            _logger = new Logger(this, this.Engine);
+            _runner = new Runner(this, deploymentLogger);
+            _logger = new EngineLogger(this, deploymentLogger, this.Engine);
         }
 
         string IDeployment.ProjectName => _projectName;
         string IDeployment.StackName => _stackName;
         bool IDeployment.IsDryRun => _isDryRun;
 
-        Serilog.ILogger IDeploymentInternal.Serilogger => _serilogger;
-        ILogger IDeploymentInternal.Logger => _logger;
+        IEngineLogger IDeploymentInternal.Logger => _logger;
         IRunner IDeploymentInternal.Runner => _runner;
 
         Stack IDeploymentInternal.Stack
@@ -173,27 +176,23 @@ namespace Pulumi
             set => Stack = value;
         }
 
-        private void InitSerilogger()
+        private ILogger CreateDefaultLogger()
         {
-            var verboseLogging = !string.IsNullOrEmpty(Environment.GetEnvironmentVariable("PULUMI_DOTNET_LOG_VERBOSE"));
-
-            var configRoot = new ConfigurationBuilder()
-                .AddEnvironmentVariables()
-                .Build();
-
-            _serilogger = new LoggerConfiguration()
-                .MinimumLevel.Is(verboseLogging ? LogEventLevel.Verbose : LogEventLevel.Fatal)
-                .ReadFrom.Configuration(configRoot)
+            var logger = new LoggerConfiguration()
+                .MinimumLevel.Is(!string.IsNullOrEmpty(Environment.GetEnvironmentVariable("PULUMI_DOTNET_LOG_VERBOSE")) ? LogEventLevel.Verbose : LogEventLevel.Fatal)
                 .WriteTo.Console()
-                .CreateLogger()
-                .ForContext<Deployment>();
+                .CreateLogger();
+
+            var loggerFactory = new SerilogLoggerFactory(logger);
+
+            return loggerFactory.CreateLogger<Deployment>();
         }
 
         private async Task<bool> MonitorSupportsFeature(string feature)
         {
             if (!this._featureSupport.ContainsKey(feature))
             {
-                var request = new Pulumirpc.SupportsFeatureRequest {Id = feature };
+                var request = new SupportsFeatureRequest {Id = feature };
                 var response = await this.Monitor.SupportsFeatureAsync(request).ConfigureAwait(false);
                 this._featureSupport[feature] = response.HasSupport;
             }
