@@ -25,7 +25,9 @@ import (
 	"github.com/blang/semver"
 	pbempty "github.com/golang/protobuf/ptypes/empty"
 	_struct "github.com/golang/protobuf/ptypes/struct"
+	"github.com/grpc-ecosystem/grpc-opentracing/go/otgrpc"
 	multierror "github.com/hashicorp/go-multierror"
+	"github.com/opentracing/opentracing-go"
 	"github.com/pkg/errors"
 	"google.golang.org/grpc/codes"
 
@@ -91,7 +93,7 @@ func NewProvider(host Host, ctx *Context, pkg tokens.Package, version *semver.Ve
 	}
 
 	plug, err := newPlugin(ctx, ctx.Pwd, path, fmt.Sprintf("%v (resource)", pkg),
-		[]string{host.ServerAddr()}, env)
+		[]string{host.ServerAddr()}, env, otgrpc.SpanDecorator(decorateProviderSpans))
 	if err != nil {
 		return nil, err
 	}
@@ -1100,11 +1102,16 @@ func (p *provider) Construct(info ConstructInfo, typ tokens.Type, name tokens.QN
 	for k, v := range info.Config {
 		config[k.String()] = v
 	}
+	configSecretKeys := []string{}
+	for _, k := range info.ConfigSecretKeys {
+		configSecretKeys = append(configSecretKeys, k.String())
+	}
 
 	resp, err := client.Construct(p.requestContext(), &pulumirpc.ConstructRequest{
 		Project:           info.Project,
 		Stack:             info.Stack,
 		Config:            config,
+		ConfigSecretKeys:  configSecretKeys,
 		DryRun:            info.DryRun,
 		Parallel:          int32(info.Parallel),
 		MonitorEndpoint:   info.MonitorAddress,
@@ -1290,6 +1297,98 @@ func (p *provider) StreamInvoke(
 	}
 }
 
+// Call dynamically executes a method in the provider associated with a component resource.
+func (p *provider) Call(tok tokens.ModuleMember, args resource.PropertyMap, info CallInfo,
+	options CallOptions) (CallResult, error) {
+	contract.Assert(tok != "")
+
+	label := fmt.Sprintf("%s.Call(%s)", p.label(), tok)
+	logging.V(7).Infof("%s executing (#args=%d)", label, len(args))
+
+	// Get the RPC client and ensure it's configured.
+	client, err := p.getClient()
+	if err != nil {
+		return CallResult{}, err
+	}
+
+	// If the provider is not fully configured, return an empty property map.
+	if !p.cfgknown {
+		return CallResult{}, nil
+	}
+
+	margs, err := MarshalProperties(args, MarshalOptions{
+		Label:         fmt.Sprintf("%s.args", label),
+		KeepUnknowns:  true,
+		KeepSecrets:   true,
+		KeepResources: true,
+	})
+	if err != nil {
+		return CallResult{}, err
+	}
+
+	// Marshal the arg dependencies.
+	argDependencies := map[string]*pulumirpc.CallRequest_ArgumentDependencies{}
+	for name, dependencies := range options.ArgDependencies {
+		urns := make([]string, len(dependencies))
+		for i, urn := range dependencies {
+			urns[i] = string(urn)
+		}
+		argDependencies[string(name)] = &pulumirpc.CallRequest_ArgumentDependencies{Urns: urns}
+	}
+
+	// Marshal the config.
+	config := map[string]string{}
+	for k, v := range info.Config {
+		config[k.String()] = v
+	}
+
+	resp, err := client.Call(p.requestContext(), &pulumirpc.CallRequest{
+		Tok:             string(tok),
+		Args:            margs,
+		ArgDependencies: argDependencies,
+		Project:         info.Project,
+		Stack:           info.Stack,
+		Config:          config,
+		DryRun:          info.DryRun,
+		Parallel:        int32(info.Parallel),
+		MonitorEndpoint: info.MonitorAddress,
+	})
+	if err != nil {
+		rpcError := rpcerror.Convert(err)
+		logging.V(7).Infof("%s failed: %v", label, rpcError.Message())
+		return CallResult{}, rpcError
+	}
+
+	// Unmarshal any return values.
+	ret, err := UnmarshalProperties(resp.GetReturn(), MarshalOptions{
+		Label:         fmt.Sprintf("%s.returns", label),
+		KeepUnknowns:  info.DryRun,
+		KeepSecrets:   true,
+		KeepResources: true,
+	})
+	if err != nil {
+		return CallResult{}, err
+	}
+
+	returnDependencies := map[resource.PropertyKey][]resource.URN{}
+	for k, rpcDeps := range resp.GetReturnDependencies() {
+		urns := make([]resource.URN, len(rpcDeps.Urns))
+		for i, d := range rpcDeps.Urns {
+			urns[i] = resource.URN(d)
+		}
+		returnDependencies[resource.PropertyKey(k)] = urns
+	}
+
+	// And now any properties that failed verification.
+	var failures []CheckFailure
+	for _, failure := range resp.GetFailures() {
+		failures = append(failures, CheckFailure{resource.PropertyKey(failure.Property), failure.Reason})
+	}
+
+	logging.V(7).Infof("%s success (#ret=%d,#failures=%d) success", label, len(ret), len(failures))
+	return CallResult{Return: ret, ReturnDependencies: returnDependencies, Failures: failures}, nil
+}
+
 // GetPluginInfo returns this plugin's information.
 func (p *provider) GetPluginInfo() (workspace.PluginInfo, error) {
 	label := fmt.Sprintf("%s.GetPluginInfo()", p.label())
@@ -1446,3 +1545,30 @@ func (ie *InitError) Error() string {
 	}
 	return err.Error()
 }
+
+func decorateSpanWithType(span opentracing.Span, urn string) {
+	if urn := resource.URN(urn); urn.IsValid() {
+		span.SetTag("pulumi-decorator", urn.Type())
+	}
+}
+
+func decorateProviderSpans(span opentracing.Span, method string, req, resp interface{}, grpcError error) {
+	if req == nil {
+		return
+	}
+
+	switch method {
+	case "/pulumirpc.ResourceProvider/Check", "/pulumirpc.ResourceProvider/CheckConfig":
+		decorateSpanWithType(span, req.(*pulumirpc.CheckRequest).Urn)
+	case "/pulumirpc.ResourceProvider/Diff", "/pulumirpc.ResourceProvider/DiffConfig":
+		decorateSpanWithType(span, req.(*pulumirpc.DiffRequest).Urn)
+	case "/pulumirpc.ResourceProvider/Create":
+		decorateSpanWithType(span, req.(*pulumirpc.CreateRequest).Urn)
+	case "/pulumirpc.ResourceProvider/Update":
+		decorateSpanWithType(span, req.(*pulumirpc.UpdateRequest).Urn)
+	case "/pulumirpc.ResourceProvider/Delete":
+		decorateSpanWithType(span, req.(*pulumirpc.DeleteRequest).Urn)
+	case "/pulumirpc.ResourceProvider/Invoke":
+		span.SetTag("pulumi-decorator", req.(*pulumirpc.InvokeRequest).Tok)
+	}
+}
