@@ -17,6 +17,7 @@ package testing
 import (
 	"bytes"
 	"fmt"
+	"io"
 	"io/ioutil"
 	"os"
 	"os/exec"
@@ -48,6 +49,11 @@ type Environment struct {
 	Backend string
 	// Environment variables to add to the environment for commands (`key=value`).
 	Env []string
+	// Passphrase for config secrets, if any
+	Passphrase string
+
+	// Content to pass on stdin, if any
+	Stdin io.Reader
 }
 
 // WriteYarnRCForTest writes a .yarnrc file which sets global configuration for every yarn inovcation. We use this
@@ -173,16 +179,24 @@ func (e *Environment) GetCommandResults(command string, args ...string) (string,
 	var outBuffer bytes.Buffer
 	var errBuffer bytes.Buffer
 
+	passphrase := "correct horse battery staple"
+	if e.Passphrase != "" {
+		passphrase = e.Passphrase
+	}
+
 	// nolint: gas
 	cmd := exec.Command(command, args...)
 	cmd.Dir = e.CWD
+	if e.Stdin != nil {
+		cmd.Stdin = e.Stdin
+	}
 	cmd.Stdout = &outBuffer
 	cmd.Stderr = &errBuffer
 	cmd.Env = os.Environ()
 	cmd.Env = append(cmd.Env, e.Env...)
 	cmd.Env = append(cmd.Env, fmt.Sprintf("%s=%s", pulumiCredentialsPathEnvVar, e.RootPath))
 	cmd.Env = append(cmd.Env, "PULUMI_DEBUG_COMMANDS=true")
-	cmd.Env = append(cmd.Env, "PULUMI_CONFIG_PASSPHRASE=correct horse battery staple")
+	cmd.Env = append(cmd.Env, fmt.Sprintf("PULUMI_CONFIG_PASSPHRASE=%s", passphrase))
 	if e.Backend != "" {
 		cmd.Env = append(cmd.Env, fmt.Sprintf("PULUMI_BACKEND_URL=%s", e.Backend))
 	}
