@@ -37,13 +37,14 @@ func construct(ctx context.Context, req *pulumirpc.ConstructRequest, engineConn
 
 	// Configure the RunInfo.
 	runInfo := RunInfo{
-		Project:     req.GetProject(),
-		Stack:       req.GetStack(),
-		Config:      req.GetConfig(),
-		Parallel:    int(req.GetParallel()),
-		DryRun:      req.GetDryRun(),
-		MonitorAddr: req.GetMonitorEndpoint(),
-		engineConn:  engineConn,
+		Project:          req.GetProject(),
+		Stack:            req.GetStack(),
+		Config:           req.GetConfig(),
+		ConfigSecretKeys: req.GetConfigSecretKeys(),
+		Parallel:         int(req.GetParallel()),
+		DryRun:           req.GetDryRun(),
+		MonitorAddr:      req.GetMonitorEndpoint(),
+		engineConn:       engineConn,
 	}
 	pulumiCtx, err := NewContext(ctx, runInfo)
 	if err != nil {
@@ -174,6 +175,7 @@ func constructInputsMap(ctx *Context, inputs map[string]interface{}) (Map, error
 	for k, v := range inputs {
 		ci := v.(*constructInput)
 
+		known := !ci.value.ContainsUnknowns()
 		value, secret, err := unmarshalPropertyValue(ctx, ci.value)
 		if err != nil {
 			return nil, errors.Wrapf(err, "unmarshaling input %s", k)
@@ -185,7 +187,7 @@ func constructInputsMap(ctx *Context, inputs map[string]interface{}) (Map, error
 		}
 
 		output := ctx.newOutput(resultType, ci.deps...)
-		output.getState().resolve(value, true /*known*/, secret, nil)
+		output.getState().resolve(value, known, secret, nil)
 		result[k] = output
 	}
 	return result, nil
@@ -216,27 +218,71 @@ func constructInputsCopyTo(ctx *Context, inputs map[string]interface{}, args int
 				continue
 			}
 
-			if field.Type.Implements(outputType) || field.Type.Implements(inputType) {
+			handleField := func(typ reflect.Type, value resource.PropertyValue, deps []Resource) (reflect.Value, error) {
 				resultType := anyOutputType
-				if field.Type.Implements(outputType) {
-					resultType = field.Type
-				} else if field.Type.Implements(inputType) {
-					toOutputMethodName := "To" + strings.TrimSuffix(field.Type.Name(), "Input") + "Output"
-					if toOutputMethod, found := field.Type.MethodByName(toOutputMethodName); found {
+				if typ.Implements(outputType) {
+					resultType = typ
+				} else if typ.Implements(inputType) {
+					toOutputMethodName := "To" + strings.TrimSuffix(typ.Name(), "Input") + "Output"
+					if toOutputMethod, found := typ.MethodByName(toOutputMethodName); found {
 						mt := toOutputMethod.Type
 						if mt.NumIn() == 0 && mt.NumOut() == 1 && mt.Out(0).Implements(outputType) {
 							resultType = mt.Out(0)
 						}
 					}
 				}
-				output := ctx.newOutput(resultType, ci.deps...)
+				output := ctx.newOutput(resultType, deps...)
 				dest := reflect.New(output.ElementType()).Elem()
-				secret, err := unmarshalOutput(ctx, ci.value, dest)
+				known := !ci.value.ContainsUnknowns()
+				secret, err := unmarshalOutput(ctx, value, dest)
+				if err != nil {
+					return reflect.Value{}, err
+				}
+				output.getState().resolve(dest.Interface(), known, secret, nil)
+				return reflect.ValueOf(output), nil
+			}
+
+			isInputType := func(typ reflect.Type) bool {
+				return typ.Implements(outputType) || typ.Implements(inputType)
+			}
+
+			if isInputType(field.Type) {
+				val, err := handleField(field.Type, ci.value, ci.deps)
 				if err != nil {
 					return err
 				}
-				output.getState().resolve(dest.Interface(), true /*known*/, secret, nil)
-				fieldV.Set(reflect.ValueOf(output))
+				fieldV.Set(val)
+				continue
+			}
+
+			if field.Type.Kind() == reflect.Slice && isInputType(field.Type.Elem()) {
+				elemType := field.Type.Elem()
+				length := len(ci.value.ArrayValue())
+				dest := reflect.MakeSlice(field.Type, length, length)
+				for i := 0; i < length; i++ {
+					val, err := handleField(elemType, ci.value.ArrayValue()[i], ci.deps)
+					if err != nil {
+						return err
+					}
+					dest.Index(i).Set(val)
+				}
+				fieldV.Set(dest)
+				continue
+			}
+
+			if field.Type.Kind() == reflect.Map && isInputType(field.Type.Elem()) {
+				elemType := field.Type.Elem()
+				length := len(ci.value.ObjectValue())
+				dest := reflect.MakeMapWithSize(field.Type, length)
+				for k, v := range ci.value.ObjectValue() {
+					key := reflect.ValueOf(string(k))
+					val, err := handleField(elemType, v, ci.deps)
+					if err != nil {
+						return err
+					}
+					dest.SetMapIndex(key, val)
+				}
+				fieldV.Set(dest)
 				continue
 			}
 
